"""Handle optional dependency imports.

Inspired from pandas: https://pandas.pydata.org/
"""

from __future__ import annotations

from importlib import import_module
from importlib.util import find_spec
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import ModuleType

# A mapping from import name to package name (on PyPI) when the package name
# is different.
_INSTALL_MAPPING: dict[str, str] = {
    "codespell_lib": "codespell",
    "cv2": "opencv-python",
    "parallel": "pyparallel",
    "pytest_cov": "pytest-cov",
    "serial": "pyserial",
    "sklearn": "scikit-learn",
    "sksparse": "scikit-sparse",
}


def import_optional_dependency(
    name: str,
    extra: str = "",
    raise_error: bool = True,
) -> ModuleType | None:
    """Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice message will be
    raised.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    raise_error : bool
        What to do when a dependency is not found.
        * True : Raise an ImportError.
        * False: Return None.

    Returns
    -------
    module : Module | None
        The imported module when found.
        None is returned when the package is not found and raise_error is False.
    """
    package_name = _INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name
    if find_spec(name) is None:
        if raise_error:
            raise ImportError(
                f"Missing optional dependency '{install_name}'. {extra} Use pip or "
                f"conda to install {install_name}."
            )
        else:
            return None
    return import_module(name)



from __future__ import annotations

from ..utils._docs import copy_doc
from ..utils.logs import _use_log_level, logger
from ._base import BaseTrigger


class MockTrigger(BaseTrigger):
    """Mock trigger class.

    Delivered triggers are logged at the 'INFO' level.
    """

    def __init__(self) -> None:
        pass

    @copy_doc(BaseTrigger.signal)
    def signal(self, value: int) -> None:
        value = super().signal(value)
        with _use_log_level("INFO"):
            logger.info("Mock set to %i.", value)



from __future__ import annotations

import click

from .sys_info import run as sys_info


@click.group()
def run() -> None:
    """Main package entry-point."""  # noqa: D401


run.add_command(sys_info)



from __future__ import annotations

import inspect
import logging
from functools import wraps
from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn_explicit

from ._checks import check_verbose
from ._docs import fill_doc
from ._fixes import WrapStdOut

if TYPE_CHECKING:
    from collections.abc import Callable


_PACKAGE: str = __package__.split(".")[0]


@fill_doc
def _init_logger(*, verbose: bool | str | int | None = None) -> logging.Logger:
    """Initialize a logger.

    Assigns sys.stdout as the first handler of the logger.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    logger : Logger
        The initialized logger.
    """
    # create logger
    verbose = check_verbose(verbose)
    logger = logging.getLogger(__package__.split(".utils", maxsplit=1)[0])
    logger.propagate = False
    logger.setLevel(verbose)
    # add the main handler
    handler = logging.StreamHandler(WrapStdOut())
    handler.setFormatter(_LoggerFormatter())
    logger.addHandler(handler)
    return logger


def add_file_handler(
    fname: str | Path,
    mode: str = "a",
    encoding: str | None = None,
    *,
    verbose: bool | str | int | None = None,
) -> None:
    """Add a file handler to the logger.

    Parameters
    ----------
    fname : str | Path
        Path to the file where the logging output is saved.
    mode : str
        Mode in which the file is opened.
    encoding : str | None
        If not None, encoding used to open the file.
    verbose : int | str | bool | None
        Sets the verbosity level of the file handler. The verbosity increases gradually
        between ``"CRITICAL"``, ``"ERROR"``, ``"WARNING"``, ``"INFO"`` and ``"DEBUG"``.
        If a bool is provided, the verbosity is set to ``"WARNING"`` for False and to
        ``"INFO"`` for True. If None is provided, the verbosity of the logger is used.

    Notes
    -----
    Don't forget to close added file handlers by iterating on ``logger.handlers`` and
    calling ``handler.close()``.
    """
    handler = logging.FileHandler(fname, mode, encoding)
    handler.setFormatter(_LoggerFormatter())
    if verbose is not None:
        verbose = check_verbose(verbose)
        handler.setLevel(verbose)
    logger.addHandler(handler)


@fill_doc
def set_log_level(verbose: bool | str | int | None) -> None:
    """Set the log level for the logger.

    Parameters
    ----------
    %(verbose)s
    """
    verbose = check_verbose(verbose)
    logger.setLevel(verbose)


class _LoggerFormatter(logging.Formatter):
    """Format string syntax."""

    # Format string syntax for the different Log levels
    _formatters = dict()
    _formatters[logging.DEBUG] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s "
        "(%(asctime)s)"
    )
    _formatters[logging.INFO] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.WARNING] = logging.Formatter(
        fmt="[%(module)s.%(funcName)s] %(levelname)s: %(message)s"
    )
    _formatters[logging.ERROR] = logging.Formatter(
        fmt="[%(module)s:%(funcName)s:%(lineno)d] %(levelname)s: %(message)s"
    )

    def __init__(self):
        super().__init__(fmt="%(levelname): %(message)s")

    def format(self, record: logging.LogRecord):  # noqa: A003
        """Format the received log record.

        Parameters
        ----------
        record : logging.LogRecord
        """
        if record.levelno <= logging.DEBUG:
            return self._formatters[logging.DEBUG].format(record)
        elif record.levelno <= logging.INFO:
            return self._formatters[logging.INFO].format(record)
        elif record.levelno <= logging.WARNING:
            return self._formatters[logging.WARNING].format(record)
        else:
            return self._formatters[logging.ERROR].format(record)


def verbose(f: Callable) -> Callable:
    """Set the verbose for the function call from the kwargs.

    Parameters
    ----------
    f : callable
        The function with a verbose argument.

    Returns
    -------
    f : callable
        The function.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        if "verbose" in kwargs:
            with _use_log_level(kwargs["verbose"]):
                return f(*args, **kwargs)
        else:
            return f(*args, **kwargs)

    return wrapper


@fill_doc
class _use_log_level:  # noqa: N801
    """Context manager to change the logging level temporary.

    Parameters
    ----------
    %(verbose)s
    """

    def __init__(self, verbose: bool | str | int | None = None):
        self._old_level = logger.level
        self._level = verbose

    def __enter__(self):
        set_log_level(self._level)

    def __exit__(self, *args):
        set_log_level(self._old_level)


def warn(
    message: str,
    category: Warning = RuntimeWarning,
    module: str = _PACKAGE,
    ignore_namespaces: tuple[str, ...] | list[str] = (_PACKAGE,),
) -> None:
    """Emit a warning with trace outside the requested namespace.

    This function takes arguments like :func:`warnings.warn`, and sends messages
    using both :func:`warnings.warn` and :func:`logging.warn`. Warnings can be
    generated deep within nested function calls. In order to provide a
    more helpful warning, this function traverses the stack until it
    reaches a frame outside the ignored namespace that caused the error.

    This function is inspired from the MNE-Python package and behaves as a smart
    'stacklevel' argument.

    Parameters
    ----------
    message : str
        Warning message.
    category : instance of Warning
        The warning class. Defaults to ``RuntimeWarning``.
    module : str
        The name of the module emitting the warning.
    ignore_namespaces : list of str | tuple of str
        Namespaces to ignore when traversing the stack.
    """
    if logging.WARNING < logger.level:
        return None
    root_dirs = [
        Path(import_module(namespace).__file__).parent
        for namespace in ignore_namespaces
    ]
    frame = inspect.currentframe()
    while frame:  # at some point it will be None and exit the loop
        fname = Path(frame.f_code.co_filename)
        if fname.parent.name == "tests":
            break  # treat tests as outside of the namespace
        lineno = frame.f_lineno
        if not (any(str(fname).startswith(str(rd)) for rd in root_dirs)):
            break
        frame = frame.f_back
    del frame
    # we need to use this instead of warn(message, category, stacklevel) because we
    # move out of our stack, so warnings won't properly recognize the module name
    # (and warnings.simplefilter will fail).
    warn_explicit(
        message,
        category,
        str(fname),
        lineno,
        module,
        globals().get("__warningregistry__", {}),
    )
    # now we emit the warning to the logger, except to the default StreamHandler on
    # stdout registered as the first handler.
    logger.handlers[0].setLevel(logging.WARNING + 1)
    logger.warning(message)
    logger.handlers[0].setLevel(0)


logger = _init_logger()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np
from matplotlib import pyplot as plt
from scipy.io import wavfile

from ..utils._checks import check_type, check_value, ensure_int, ensure_path
from ..utils._docs import copy_doc, fill_doc
from ..utils.logs import logger, warn
from .backend import BACKENDS
from .backend._base import BaseBackend

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
    from numpy.typing import NDArray

    from ..time import BaseClock


@fill_doc
class BaseSound(ABC):
    """Base audio stimulus class.

    Parameters
    ----------
    %(audio_volume)s
    %(audio_duration)s
    %(audio_sample_rate)s
    %(audio_device)s
    %(audio_n_channels)s
    %(audio_backend)s
    %(audio_clock)s
    %(audio_kwargs)s
    """

    @abstractmethod
    def __init__(
        self,
        volume: float | Sequence[float],
        duration: float,
        sample_rate: int | None,
        device: int | None,
        n_channels: int = 1,
        *,
        backend: str,
        clock: BaseClock,
        **kwargs,
    ) -> None:
        check_type(backend, (str,), "backend")
        check_value(backend, BACKENDS, "backend")
        _check_duration(duration)
        self._duration = duration
        self._n_channels = ensure_int(n_channels, "n_channels")
        if self._n_channels < 1:
            raise ValueError(
                "The number of channels must be at least 1. Provided "
                f"'{self._n_channels}' is invalid."
            )
        self._volume = _ensure_volume(volume, self._n_channels)
        # window can only be set by property
        self._window = None
        # the arguments sample_rate, device, clock, and **kwargs are checked in
        # the backend initialization.
        self._backend_kwargs = kwargs
        self._backend = BACKENDS[backend](device, sample_rate, clock=clock)
        self._set_times()
        self._set_signal()

    def _set_times(self) -> None:
        """Set the timestamp array."""
        self._times = np.linspace(
            0,
            self.duration,
            int(self.duration * self.sample_rate),
            endpoint=False,
            dtype=np.float32,
        )

    @abstractmethod
    def _set_signal(self, signal: NDArray) -> None:
        """Set the signal array."""
        if self._window is not None:
            assert self._window.size == signal.shape[0]  # sanity-check
            signal = signal * self._window[:, np.newaxis]
        assert self._volume.ndim == 1  # sanity-check
        assert self._volume.size == self._n_channels  # sanity-check
        signal = np.ascontiguousarray(signal * self._volume / 100, dtype=np.float32)
        self._signal = signal

    @copy_doc(BaseBackend.play)
    def play(self, when: float | None = None, *, blocking: bool = False) -> None:
        self._backend.play(when=when, blocking=blocking)

    def plot(self) -> tuple[Figure, Axes]:
        """Plot the audio signal waveform.

        Returns
        -------
        f : Figure
            The matplotlib figure object.
        ax : Axes
            The matplotlib axes object.
        """
        f, ax = plt.subplots(self._n_channels, 1, layout="constrained")
        if self._n_channels == 1:
            ax = np.array([ax])
        for ch in range(self._n_channels):
            ax[ch].plot(self.times, self.signal[:, ch])
            ax[ch].set_ylabel("Amplitude")
            ax[ch].set_title(repr(self))
        ax[-1].set_xlabel("Time [s]")
        return f, ax if ax.size != 1 else ax[0]

    def save(self, fname: str | Path, *, overwrite: bool = False) -> None:
        """Save the audio stimulus to a WAV file.

        The saving is handled by :func:`scipy.io.wavfile.write`.

        Parameters
        ----------
        fname : str | Path
            Path to the output file. The extension should be ``'.wav'``.
        overwrite : bool
            If True, existing files are overwritten.
        """
        fname = ensure_path(fname, must_exist=False)
        if fname.suffix != ".wav":
            raise ValueError("The file extension must be '.wav'.")
        if overwrite is False and fname.exists():
            raise FileExistsError(
                f"The file '{fname}' already exists. Set 'overwrite' to True."
            )
        fname.parent.mkdir(parents=True, exist_ok=True)
        logger.debug("Saving sound to %s.", fname)
        wavfile.write(fname, self.sample_rate, self.signal)

    @copy_doc(BaseBackend.stop)
    def stop(self) -> None:
        self._backend.stop()

    @property
    def duration(self) -> float:
        """The duration of the audio stimulus.

        :type: :class:`float`
        """
        return self._duration

    @duration.setter
    def duration(self, duration: float):
        logger.debug("Setting 'duration' to %s [s].", duration)
        _check_duration(duration)
        self._duration = duration
        self._set_times()
        self._set_signal()

    @property
    def sample_rate(self) -> int:
        """The sample rate of the audio stimulus.

        :type: :class:`int`
        """
        return self._backend.sample_rate

    @property
    def signal(self) -> NDArray[np.float32]:
        """The audio signal.

        :type: :class:`~numpy.ndarray` of shape (n_samples, n_channels)
        """
        return self._signal

    @property
    def times(self) -> NDArray[np.float32]:
        """The time array of the audio stimulus.

        :type: :class:`~numpy.ndarray` of shape (n_samples,)
        """
        return self._times

    @property
    def volume(self) -> NDArray[np.float32]:
        """The volume of the audio stimulus per channel given as a percentage.

        :type: :class:`~numpy.ndarray` of shape (n_channels,)
        """
        return self._volume

    @volume.setter
    def volume(self, volume: float | Sequence[float]) -> None:
        logger.debug("Setting 'volume' to %s [%].", volume)
        self._volume = _ensure_volume(volume, self._n_channels)
        self._set_signal()

    @property
    def window(self) -> NDArray[np.float32] | None:
        """Window applied to the audio signal.

        :type: :class:`~numpy.ndarray` of shape (n_samples,) | None
        """
        return self._window

    @window.setter
    def window(self, window: NDArray | list | tuple | set | None) -> None:
        self._window = _ensure_window(window, self.times.size)
        self._set_signal()

    @property
    def _signal(self) -> NDArray[np.float32]:
        """The audio signal."""
        return self._signal_array

    @_signal.setter
    def _signal(self, signal: NDArray[np.float32]) -> None:
        self._backend.close()  # np-op if already closed
        self._signal_array = signal
        self._backend.initialize(self._signal_array, **self._backend_kwargs)


def _ensure_volume(volume: float | Sequence[float], n_channels: int) -> None:
    """Check that the volume is valid."""
    if isinstance(volume, (float | int)):
        volume = np.full(n_channels, volume, dtype=np.float32)
    check_type(volume, ("array-like",), "volume")
    volume = np.asarray(volume, dtype=np.float32)
    if volume.ndim != 1:
        raise ValueError(
            "The volume must be a single value or a sequence of values. Provided "
            f"'{volume}' is invalid."
        )
    if volume.size != n_channels:
        raise ValueError(
            "The number of volume values must match the number of channels. "
            f"Provided '{volume.size}' values for '{n_channels}' channels."
        )
    if np.any(volume < 0) or np.any(100 < volume):
        raise ValueError(
            "The volume must be a percentage between 0 and 100. Provided "
            f"'{volume}' is invalid."
        )
    return np.asarray(volume, dtype=np.float32)


def _ensure_window(window: NDArray | list | tuple | set | None, n_samples: int) -> None:
    """Check that the window is valid."""
    if window is not None:
        check_type(window, ("array-like",), "window")
        window = np.asarray(window, dtype=np.float32)
        if window.ndim != 1:
            raise ValueError(
                f"The window must be a 1D array. Provided'{window.ndim}' dimensions "
                "are invalid."
            )
        if window.size != n_samples:
            raise ValueError(
                "The window must have the same number of samples as the signal. "
                f"Provided '{window.size}' samples for '{n_samples}' samples."
            )
        if np.any(window < 0) or np.any(1 < window):
            warn("The window values should be normalized between 0 and 1.")
            window = (window - np.min(window)) / (np.max(window) - np.min(window))
    return window


def _check_duration(duration: float) -> None:
    """Check if the duration is valid."""
    check_type(duration, ("numeric",), "duration")
    if duration <= 0:
        raise ValueError(
            "The argument 'duration' must be a strictly positive number defining "
            f"the length of the sound in seconds. Provided '{duration}' is invalid."
        )



from __future__ import annotations

import time
from typing import TYPE_CHECKING

from ._clock import Clock

if TYPE_CHECKING:
    from . import BaseClock


def sleep(duration: float, *, clock: BaseClock = Clock) -> None:
    """High precision sleep function.

    Parameters
    ----------
    duration : float
        Duration to sleep in seconds. If the value is less than or equal to 0, the
        function returns immediately.
    clock : BaseClock
        Clock object to use for time measurement. By default, the
        :class:`stimuli.time.Clock` class is used.

    Notes
    -----
    On Windows, only python version 3.11 and above have good accuracy when sleeping.
    This accuracy limitation dictates the minimum version of ``stimuli``.
    """
    if duration <= 0:
        return
    clock = clock()
    duration = int(duration * 1e9)
    while True:
        remaining_time = duration - clock.get_time_ns()  # nanoseconds
        if remaining_time <= 0:
            break
        if remaining_time >= 200000:  # 200 microseconds
            time.sleep(remaining_time * 1e-9 / 2)



"""Temporary bug-fixes awaiting an upstream fix."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


# https://github.com/sphinx-gallery/sphinx-gallery/issues/1112
class WrapStdOut:
    """Dynamically wrap to sys.stdout.

    This makes packages that monkey-patch sys.stdout (e.g.doctest, sphinx-gallery) work
    properly.
    """

    def __getattr__(self, name: str) -> Any:  # noqa: D105
        # Even more ridiculous than this class, this must be sys.stdout (not
        # just stdout) in order for this to work (tested on OSX and Linux).
        if hasattr(sys.stdout, name):
            return getattr(sys.stdout, name)
        else:
            raise AttributeError(f"'file' object has not attribute '{name}'.")



from __future__ import annotations

import platform
import sys
from functools import partial
from importlib.metadata import metadata, requires, version
from typing import TYPE_CHECKING

import psutil
import sounddevice as sd
from packaging.requirements import Requirement

from ._checks import check_type

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import IO


def sys_info(fid: IO | None = None, developer: bool = False):
    """Print the system information for debugging.

    Parameters
    ----------
    fid : file-like | None
        The file to write to, passed to :func:`print`. Can be None to use
        :data:`sys.stdout`.
    developer : bool
        If True, display information about optional dependencies.
    """
    check_type(developer, (bool,), "developer")

    ljust = 26
    out = partial(print, end="", file=fid)
    package = __package__.split(".")[0]

    # OS information - requires python 3.8 or above
    out("Platform:".ljust(ljust) + platform.platform() + "\n")
    # python information
    out("Python:".ljust(ljust) + sys.version.replace("\n", " ") + "\n")
    out("Executable:".ljust(ljust) + sys.executable + "\n")
    # CPU information
    out("CPU:".ljust(ljust) + platform.processor() + "\n")
    out("Physical cores:".ljust(ljust) + str(psutil.cpu_count(False)) + "\n")
    out("Logical cores:".ljust(ljust) + str(psutil.cpu_count(True)) + "\n")
    # memory information
    out("RAM:".ljust(ljust))
    out(f"{psutil.virtual_memory().total / float(2 ** 30):0.1f} GB\n")
    out("SWAP:".ljust(ljust))
    out(f"{psutil.swap_memory().total / float(2 ** 30):0.1f} GB\n")
    # package information
    out(f"{package}:".ljust(ljust) + version(package) + "\n")

    # dependencies
    out("\nCore dependencies\n")
    dependencies = [Requirement(elt) for elt in requires(package)]
    core_dependencies = [dep for dep in dependencies if "extra" not in str(dep.marker)]
    _list_dependencies_info(out, ljust, package, core_dependencies)

    # extras
    if developer:
        keys = sorted(
            [
                elt
                for elt in metadata(package).get_all("Provides-Extra")
                if elt not in ("all", "full")
            ]
        )
        for key in keys:
            extra_dependencies = [
                dep
                for dep in dependencies
                if all(elt in str(dep.marker) for elt in ("extra", key))
            ]
            if len(extra_dependencies) == 0:
                continue
            out(f"\nOptional '{key}' dependencies\n")
            _list_dependencies_info(out, ljust, package, extra_dependencies)

    # audio device information
    out("\nAudio devices\n")
    out(repr(sd.query_devices()) + "\n")


def _list_dependencies_info(
    out: Callable, ljust: int, package: str, dependencies: list[Requirement]
) -> None:
    """List dependencies names and versions."""
    unicode = sys.stdout.encoding.lower().startswith("utf")
    if unicode:
        ljust += 1

    not_found: list[Requirement] = list()
    for dep in dependencies:
        if dep.name == package:
            continue
        try:
            version_ = version(dep.name)
        except Exception:
            not_found.append(dep)
            continue

        # build the output string step by step
        output = f"✔︎ {dep.name}" if unicode else dep.name
        # handle version specifiers
        if len(dep.specifier) != 0:
            output += f" ({str(dep.specifier)})"
        output += ":"
        output = output.ljust(ljust) + version_

        # handle special dependencies with backends, C dep, ..
        if dep.name in ("matplotlib", "seaborn") and version_ != "Not found.":
            try:
                from matplotlib import pyplot as plt

                backend = plt.get_backend()
            except Exception:
                backend = "Not found"

            output += f" (backend: {backend})"
        out(output + "\n")

    if len(not_found) != 0:
        not_found = [
            f"{dep.name} ({str(dep.specifier)})"
            if len(dep.specifier) != 0
            else dep.name
            for dep in not_found
        ]
        if unicode:
            out(f"✘ Not installed: {', '.join(not_found)}\n")
        else:
            out(f"Not installed: {', '.join(not_found)}\n")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from ...time import BaseClock
from ...utils._checks import check_type
from ...utils.logs import warn

if TYPE_CHECKING:
    from numpy.typing import NDArray


class BaseBackend(ABC):
    """Base backend class for audio stimuli.

    Parameters
    ----------
    device : int | None
        Device identifier. If None, the default output device is used.
    sample_rate : int
        The sample rate of the audio data, which should match the sample rate of the
        output device. If None, the default sample rate of the device is used.
    clock : BaseClock
        Clock object used for time measurement. By default, the
        :class:`stimuli.time.Clock` class is used.
    """

    @abstractmethod
    def __init__(
        self,
        device: int | None,
        sample_rate: int | None,
        clock: BaseClock,
    ) -> None:
        self._clock = clock()
        check_type(self._clock, (BaseClock,), "clock")

    def __del__(self) -> None:
        """Make sure that we kill the stream during deletion."""
        self.close()  # no-op if already closed

    @abstractmethod
    def close(self) -> None:
        """Close the backend and release the resources."""

    @abstractmethod
    def initialize(self, data: NDArray) -> None:
        """Initialize the backend with audio data.

        Parameters
        ----------
        data : :class:`~numpy.ndarray` of shape (n_frames, n_channels)
            The audio data to play provided as a 2 dimensional array of shape
            ``(n_frames, n_channels)``. The array layout must be C-contiguous. A one
            dimensional :class:`~numpy.ndarray` of shape ``(n_frames,)`` is also
            accepted for mono audio.
        """
        check_type(data, (np.ndarray,), "data")
        if data.ndim not in (1, 2):
            raise ValueError(
                "The data array must be 1D or 2D of shape (n_frames, n_channels). "
                f"The provided array has {data.ndim} dimensions."
            )
        if not data.flags["C_CONTIGUOUS"]:
            warn(
                "The data array provided to the 'SoundSD' backend is not C-contiguous."
            )
            data = np.ascontiguousarray(data)
        self._data = data if data.ndim == 2 else data[:, np.newaxis]
        self._duration = self._data.shape[0] / self._sample_rate

    @abstractmethod
    def play(self, when: float | None = None, *, blocking: bool = False) -> None:
        """Play the audio data.

        Parameters
        ----------
        when : float | None
            The relative time in seconds when to start playing the audio data. For
            instance, ``0.2`` will start playing in 200 ms. If ``None``, the audio data
            is played as soon as possible. A duration superior to the device latency is
            recommended.
        blocking : bool
            If True, the function blocks until the audio playback is finished.
        """
        self._check_initialized()
        if when is not None:  # quick input validation
            assert 0 < when, "The 'when' parameter must be a positive float."

    @abstractmethod
    def stop(self) -> None:
        """Interrupt immediately the playback of the audio data."""
        self._check_initialized()

    def _check_initialized(self) -> None:
        """Check if the backend is initialized with audio data."""
        if not hasattr(self, "_data"):
            raise RuntimeError("The backend is not initialized with sound data.")

    @property
    def clock(self) -> BaseClock:
        """The clock object used for time measurement."""
        return self._clock

    @property
    def duration(self) -> float:
        """The duration of the audio data, in seconds."""
        return self._duration

    @property
    def sample_rate(self) -> int:
        """The sample rate of the audio data."""
        return self._sample_rate



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..time import Clock
from ..utils._checks import check_type, check_value
from ..utils._docs import copy_doc, fill_doc
from ..utils.logs import logger
from ._base import BaseSound

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ..time import BaseClock


@fill_doc
class SoundAM(BaseSound):
    """Amplitude modulated sound.

    This sound is composed of a carrier frequency ``fc`` which is amplitude modulated at
    ``fm``.

    Parameters
    ----------
    frequency_carrier : float
        Carrier frequency in Hz.
    frequency_modulation : float
        Modulation frequency in Hz.
    method : ``'conventional'`` | ``'dsbsc'``
        ``'conventional'`` is also called classical AM, the equation used is::

                signal = (1 - M(t)) * cos(2*pi*fc*t)
                M(t) = cos(2*pi*fm*t)

        ``'dsbsc'`` is also called double side band suppressed carrier, the equation
        used is::

                signal = M(t)*cos(2*pi*fc*t)
                M(t) = sin(2*pi*fm*t)
    %(audio_volume)s
    %(audio_duration)s
    %(audio_sample_rate)s
    %(audio_device)s
    %(audio_n_channels)s
    %(audio_backend)s
    %(audio_clock)s
    %(audio_kwargs)s
    """

    def __init__(
        self,
        frequency_carrier: float,
        frequency_modulation: float,
        method: str,
        volume: float | Sequence[float],
        duration: float,
        sample_rate: int | None = None,
        device: int | None = None,
        n_channels: int = 1,
        *,
        backend: str = "sounddevice",
        clock: BaseClock = Clock,
        **kwargs,
    ) -> None:
        _check_frequency(frequency_carrier, "carrier")
        self._frequency_carrier = frequency_carrier
        _check_frequency(frequency_modulation, "modulation")
        self._frequency_modulation = frequency_modulation
        check_type(method, (str,), "method")
        method = method.lower().strip()
        check_value(method, ("conventional", "dsbsc"), "AM method")
        self._method = method
        super().__init__(
            volume,
            duration,
            sample_rate,
            device,
            n_channels,
            backend=backend,
            clock=clock,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Representation of the object."""
        return (
            f"<{self._method}-AM sound @ {self._frequency_carrier:.2f} % "
            f"{self._frequency_modulation:.2f} Hz - {self.duration:.2f} s>"
        )

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        if self._method == "conventional":
            amplitude = 1 - np.cos(2 * np.pi * self._frequency_modulation * self._times)
            signal = amplitude * np.cos(
                2 * np.pi * self._frequency_carrier * self._times
            )

        elif self._method == "dsbsc":
            amplitude = np.sin(2 * np.pi * self._frequency_modulation * self._times)
            signal = amplitude * np.sin(
                2 * np.pi * self._frequency_carrier * self._times
            )
        signal /= np.max(np.abs(signal))  # normalize
        signal = np.vstack([signal] * self._n_channels).T
        super()._set_signal(signal)

    @property
    def frequency_carrier(self) -> float:
        """Sound's carrier frequency in Hz.

        :type: :class:`float`
        """
        return self._frequency_carrier

    @frequency_carrier.setter
    def frequency_carrier(self, frequency_carrier: float) -> None:
        logger.debug("Setting 'frequency_carrier' to %.2f [Hz].", frequency_carrier)
        _check_frequency(frequency_carrier, "carrier")
        self._frequency_carrier = frequency_carrier
        self._set_signal()

    @property
    def frequency_modulation(self) -> float:
        """Sound's modulation frequency in Hz.

        :type: :class:`float`
        """
        return self._frequency_modulation

    @frequency_modulation.setter
    def frequency_modulation(self, frequency_modulation: float) -> None:
        logger.debug(
            "Setting 'frequency_modulation' to %.2f [Hz].", frequency_modulation
        )
        _check_frequency(frequency_modulation, "modulation")
        self._frequency_modulation = frequency_modulation
        self._set_signal()

    @property
    def method(self) -> str:
        """The amplitude modulation method.

        :type: :class:`str`
        """
        return self._method

    @method.setter
    def method(self, method: str) -> None:
        logger.debug("Setting 'method' to %s.", method)
        check_type(method, (str,), "method")
        method = method.lower().strip()
        check_value(method, ("conventional", "dsbsc"), "AM method")
        self._method = method
        self._set_signal()


def _check_frequency(frequency: float, ftype: str) -> None:
    """Check that the given frequency is valid."""
    check_type(frequency, ("numeric",), item_name=f"frequency_{ftype}")
    if frequency <= 0:
        raise ValueError(
            f"The {ftype} frequency must be a strictly positive number. Provided "
            f"'{frequency}' is invalid."
        )



from __future__ import annotations

from abc import ABC, abstractmethod


class BaseTrigger(ABC):
    """Base trigger class."""

    @abstractmethod
    def __init__(self):  # pragma: no cover
        pass

    @abstractmethod
    def signal(self, value: int) -> int:
        """Send a trigger value.

        Parameters
        ----------
        value : int
            Value of the trigger, between 1 and 255.
        """
        try:
            value = int(value)
        except TypeError:
            raise TypeError(
                "The argument 'value' of a trigger must be an integer "
                "between 1 and 255 included."
            )
        if not (1 <= value <= 255):
            raise ValueError(
                "The argument 'value' of a trigger must be an integer "
                "between 1 and 255 included."
            )
        return value



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..time import Clock
from ..utils._checks import check_type, check_value
from ..utils._docs import copy_doc, fill_doc
from ..utils.logs import logger
from ._base import BaseSound

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence

    from ..time import BaseClock


_PSDS: dict[str, Callable] = {
    "white": lambda f: 1,
    "blue": lambda f: np.sqrt(f),
    "violet": lambda f: f,
    "pink": lambda f: 1 / np.where(f == 0, np.inf, np.sqrt(f)),
    "brown": lambda f: 1 / np.where(f == 0, np.inf, f),
}


@fill_doc
class Noise(BaseSound):
    """Colored noise stimulus.

    Parameters
    ----------
    color : ``'white'`` | ``'pink'`` | ``'blue'`` | ``'violet'`` | ``'brown'``
        The name of the noise color.
    %(audio_volume)s
    %(audio_duration)s
    %(audio_sample_rate)s
    %(audio_device)s
    %(audio_n_channels)s
    %(audio_backend)s
    %(audio_clock)s
    %(audio_kwargs)s
    """

    def __init__(
        self,
        color: str,
        volume: float | Sequence[float],
        duration: float,
        sample_rate: int | None = None,
        device: int | None = None,
        n_channels: int = 1,
        *,
        backend: str = "sounddevice",
        clock: BaseClock = Clock,
        **kwargs,
    ) -> None:
        check_type(color, (str,), "color")
        color = color.lower().strip()
        check_value(color, _PSDS, "color")
        self._color = color
        self._rng = np.random.default_rng()
        super().__init__(
            volume,
            duration,
            sample_rate,
            device,
            n_channels,
            backend=backend,
            clock=clock,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Representation of the object."""
        return f"<{self._color.capitalize()} noise @ {self.duration:.2f} s>"

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        white = self._rng.standard_normal(size=self._times.size, dtype=np.float32)
        dft = np.fft.rfft(white)
        S = _PSDS[self._color](np.fft.rfftfreq(self._times.size))
        # nornalize to preserve the energy from the white noise
        S /= np.sqrt(np.mean(S**2))
        signal = np.fft.irfft(dft * S)
        signal /= np.max(np.abs(signal))  # normalize
        signal = np.vstack([signal] * self._n_channels).T
        super()._set_signal(signal)
        # make sure we have the correct times as the rFFT and irFFT could get us off
        if self._times.size != self._signal.size:
            self._times = self._times[: self._signal.size]
            self._duration = self._times[-1] + 1 / self.sample_rate

    @property
    def color(self) -> str:
        """The color of the noise.

        :type: :class:`str`
        """
        return self._color

    @color.setter
    def color(self, color: str) -> None:
        logger.debug("Setting 'color' to %s.", color)
        check_type(color, (str,), "color")
        color = color.lower().strip()
        check_value(color, _PSDS, "color")
        self._color = color
        self._set_signal()



from __future__ import annotations

import click

from .. import sys_info


@click.command(name="sys-info")
@click.option(
    "--developer",
    help="Display information for optional dependencies.",
    is_flag=True,
)
def run(developer: bool) -> None:
    """Run sys_info() command."""
    sys_info(developer=developer)



from __future__ import annotations

import time
from abc import ABC, abstractmethod

from ..utils._docs import copy_doc


class BaseClock(ABC):
    """Base class for high precision clocks.

    If you want to implement a custom clock, you should subclass this class and define
    the abstract method :meth:`stimuli.time.BaseClock.get_time_ns` which should return
    the time elapsed since instantiation in nanoseconds.
    """

    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_time_ns(self) -> int:
        """Return the current time in nanoseconds.

        Returns
        -------
        time : int
            The current time in nanoseconds.
        """

    def get_time_us(self) -> float:
        """Return the current time in microseconds.

        Returns
        -------
        time : float
            The current time in microseconds.
        """
        return self.get_time_ns() / 1e3

    def get_time_ms(self) -> float:
        """Return the current time in milliseconds.

        Returns
        -------
        time : float
            The current time in milliseconds.
        """
        return self.get_time_ns() / 1e6

    def get_time(self) -> float:
        """Return the current time in seconds.

        Returns
        -------
        time : float
            The current time in seconds.
        """
        return self.get_time_ns() / 1e9


class Clock(BaseClock):
    """Clock which keeps track of time in nanoseconds.

    The origin ``t=0`` corresponds to the instantiation of the
    :class:`stimuli.time.Clock` object. The time is measured either through the
    monotonic :func:`time.monotonic_ns` function or through the performance counter
    :func:`time.perf_counter` function, depending on which one has the highest
    resolution.
    """

    def __init__(self) -> None:
        if (
            time.get_clock_info("perf_counter").resolution
            < time.get_clock_info("monotonic").resolution
        ):
            self._function = time.perf_counter_ns
        else:
            self._function = time.monotonic_ns
        self._t0 = self._function()

    @copy_doc(BaseClock.get_time_ns)
    def get_time_ns(self) -> int:
        return self._function() - self._t0



from __future__ import annotations

import time
from concurrent.futures import ThreadPoolExecutor
from platform import system

from ..utils._checks import check_type, check_value, ensure_int
from ..utils._docs import copy_doc
from ..utils._imports import import_optional_dependency
from ..utils.logs import logger
from ._base import BaseTrigger


class ParallelPortTrigger(BaseTrigger):
    """Trigger using a parallel port (also called LPT port).

    Parameters
    ----------
    address : int (hex) | str
        The address of the parallel port on the system.
        If an :ref:`api/trigger:Arduino to Parallel Port converter` is
        used, the address must be the serial port address or ``"arduino"`` for automatic
        detection.
    port_type : str | None
        Either ``'arduino'`` or ``'pport'`` depending on the connection.
        If None, attempts to infers the type of port from the address.
    delay : int
        Delay in milliseconds until which a new trigger cannot be sent. During
        this time, the pins of the LPT port remain in the same state.

    Notes
    -----
    The address is specific to the system. Typical parallel port addresses are:

    - On Linux::

          LPT1 = /dev/parport0
          LPT2 = /dev/parport1
          LPT3 = /dev/parport2

    - On Windows::

          LPT1 = 0x0378 or 0x03BC
          LPT2 = 0x0278 or 0x0378
          LPT3 = 0x0278

    - macOS does not have support for built-in parallel ports.
    """

    def __init__(
        self, address: int | str, port_type: str | None = None, delay: int = 10
    ) -> None:
        check_type(address, ("int-like", str), "address")
        if not isinstance(address, str):
            address = ensure_int(address)
        delay = ensure_int(delay, "delay")
        self._delay = delay / 1000.0
        if port_type is None:
            self._port_type = ParallelPortTrigger._infer_port_type(address)
        else:
            check_type(port_type, (str,), "port_type")
            check_value(port_type, ("arduino", "pport"), "port_type")
            self._port_type = port_type

        # initialize port
        if self._port_type == "arduino":
            import_optional_dependency(
                "serial", extra="Install 'pyserial' for ARDUINO support."
            )
            if address == "arduino":
                self._address = ParallelPortTrigger._search_arduino()
            else:
                self._address = address
            self._connect_arduino()

        elif self._port_type == "pport":
            if system() == "Linux":
                import_optional_dependency(
                    "parallel",
                    extra="Install 'pyparallel' for LPT support on Linux.",
                )
            self._address = address
            self._connect_pport()
        # set pins to 0 and prepare threadpool for resets
        self._set_data(0)
        self._executor = ThreadPoolExecutor(max_workers=1)
        self._future = None

    @staticmethod
    def _infer_port_type(address: int | str) -> str:
        """Infer the type of port from the address."""
        if system() == "Linux":
            if not isinstance(address, str):
                raise TypeError(
                    "On Linux, a parallel port address must be provided as a string."
                )
            if address.startswith("/dev/parport"):
                return "pport"
            elif address.startswith("/dev/ttyACM") or address == "arduino":
                return "arduino"
            else:
                raise RuntimeError(
                    f"Could not infer the port type from the address '{address}'. "
                    "Please provide the 'port_type' argument when creating the "
                    "ParallelPortTrigger object."
                )
        elif system() == "Darwin":
            if address == "arduino":
                return "arduino"
            else:
                raise RuntimeError(
                    "macOS does not support on-board parallel ports. Only arduino "
                    "converters are supported with address='arduino'."
                )
        elif system() == "Windows":
            if isinstance(address, int):
                return "pport"
            elif address.startswith("COM") or address == "arduino":
                return "arduino"
            else:
                raise RuntimeError(
                    f"Could not infer the port type from the address '{address}'. "
                    "Please provide the 'port_type' argument when creating the "
                    "ParallelPortTrigger object."
                )

    @staticmethod
    def _search_arduino() -> str:
        """Look for a connected Arduino to LPT converter."""
        from serial.tools import list_ports

        for arduino in list_ports.grep(regexp="Arduino"):
            logger.info("Found arduino to LPT on '%s'.", arduino)
            return arduino.device
        else:
            raise OSError("No arduino card was found.")

    def _connect_arduino(self, baud_rate: int = 115200) -> None:
        """Connect to an Arduino to LPT converter."""
        from serial import Serial, SerialException

        try:
            self._port = Serial(self._address, baud_rate)
        except SerialException:
            msg = f"Could not access arduino to LPT on '{self._address}'."
            if system() == "Linux":
                msg += (
                    " Make sure you have the permission to access this "
                    "address, e.g. by adding your user account to the "
                    "'dialout' group: 'sudo usermod -a -G dialout <username>'."
                )
            raise SerialException(msg)

        time.sleep(1)
        logger.info("Connected to arduino to LPT on '%s'.", self._address)

    def _connect_pport(self) -> None:
        """Connect to the ParallelPort."""
        from .io import ParallelPort

        if ParallelPort is None and system() == "Darwin":
            raise RuntimeError(
                "macOS does not support built-in parallel port. Please use an arduino "
                "to LPT converter for hardware triggers or stimuli.trigger.LSLTrigger "
                "for software triggers."
            )
        elif ParallelPort is None and system() != "Linux":
            raise RuntimeError(
                "Windows supports built-in parallel port via inpout32, inpout64 or "
                "dlportio. Neither of this driver was found."
            )

        try:
            self._port = ParallelPort(self._address)
        except Exception:
            msg = f"Could not access the parallel port on '{self._address}'."
            if system() == "Linux":
                msg += (
                    " Make sure you have the permission to access this "
                    "address, e.g. by adding your user account to the 'lp' "
                    "group: 'sudo usermod -a -G lp <username>'. Make sure the "
                    "'lp' module is removed and the 'ppdev' module is loaded: "
                    "'sudo rmmod lp' & 'sudo modprobe ppdev'. You can "
                    "configure the module loaded by default in '/etc/modprobe.d/'."
                )
            raise RuntimeError(msg)

        time.sleep(1)
        logger.info("Connected to parallel port on '%s'.", self._address)

    @copy_doc(BaseTrigger.signal)
    def signal(self, value: int) -> None:
        value = super().signal(value)
        if self._future is not None and not self._future.done():
            logger.warning(
                "You are sending a new signal before the end of the "
                "last signal. Signal ignored. Delay required = %.1f ms.",
                self.delay,
            )
        self._set_data(value)
        self._future = self._executor.submit(self._signal_off())

    def _signal_off(self) -> None:
        """Reset trigger signal to 0."""
        time.sleep(self._delay)
        self._set_data(0)

    def _set_data(self, value: int) -> None:
        """Set data on the pin."""
        if self._port_type == "arduino":
            self._port.write(bytes([value]))
        else:
            self._port.setData(value)

    def close(self) -> None:
        """Disconnect the parallel port.

        This method should free the parallel or serial port and let other application or
        python process use it.
        """
        if hasattr(self, "_executor"):
            self._executor.shutdown(wait=True, cancel_futures=False)
        if (
            hasattr(self, "_port_type")
            and self._port_type == "arduino"
            and hasattr(self, "_port")
        ):
            try:
                self._port.close()
            except Exception:
                pass
            try:
                del self._port
            except Exception:
                pass

    def __del__(self) -> None:  # noqa: D105
        self.close()

    # --------------------------------------------------------------------
    @property
    def address(self) -> int | str:
        """The address of the parallel port on the system.

        :type: int | str
        """
        return self._address

    @property
    def delay(self) -> float:
        """Delay (ms) to wait between two :meth:`~ParallelPortTrigger.signal`.

        :type: float
        """
        return self._delay * 1000.0

    @property
    def port_type(self) -> str:
        """Type of connection port.

        :type: str
        """
        return self._port_type



# We deliberately delay importing the pyparallel module until we try
# to use it - this allows us to import the class on machines
# which don't have it and then worry about dealing with
# using the right one later

# This is necessary to stop the local parallel.py masking the module
# we actually want to find!

# We duck-type the parallel port objects


class PParallelLinux:
    """Clsas for read/write access to the parallel port on a Linux.

    Uses pyparallel.

    Note that you must have the lp module removed and the ppdev module loaded
    to use this code::

        sudo rmmod lp
        sudo modprobe ppdev
    """

    def __init__(self, address="/dev/parport0"):
        """Set the device node of your parallel port.

        Common port addresses::

            LPT1 = /dev/parport0
            LPT2 = /dev/parport1
            LPT3 = /dev/parport2
        """
        import parallel as pyp

        try:
            self.port = pyp.Parallel(address)
        except FileNotFoundError:
            raise RuntimeError(
                f"Could not access parallel port on '{address}'. "
                "No such file or directory."
            )
        self.status = None

    def __del__(self):
        """Delete and free the port."""
        if hasattr(self, "port"):
            del self.port

    def setData(self, data):
        """Set the data to be presented on the parallel port (one ubyte).

        Alternatively you can set the value of each pin (data pins are pins
        2-9 inclusive) using :func:`~psychopy.parallel.setPin`

        Examples::

            p.setData(0)  # sets all pins low
            p.setData(255)  # sets all pins high
            p.setData(2)  # sets just pin 3 high (remember that pin2=bit0)
            p.setData(3)  # sets just pins 2 and 3 high

        You can also convert base 2 to int easily in python::

            parallel.setData(int("00000011", 2))  # pins 2 and 3 high
            parallel.setData(int("00000101", 2))  # pins 2 and 4 high
        """
        self.port.setData(data)

    def setPin(self, pinNumber, state):
        """Set a desired pin to be high(1) or low(0).

        Only pins 2-9 (incl) are normally used for data output::

            p.setPin(3, 1)  # sets pin 3 high
            p.setPin(3, 0)  # sets pin 3 low
        """
        # I can't see how to do this without reading and writing the data
        if state:
            self.port.setData(self.port.PPRDATA() | (2 ** (pinNumber - 2)))
        else:
            self.port.setData(self.port.PPRDATA() & (255 ^ 2 ** (pinNumber - 2)))

    def readData(self):
        """Return the value currently set on the data pins (2-9)."""
        return self.port.PPRDATA()

    def readPin(self, pinNumber):
        """Determine whether a desired (input) pin is high(1) or low(0).

        Pins 2-13 and 15 are currently read here
        """
        if pinNumber == 10:
            return self.port.getInAcknowledge()
        elif pinNumber == 11:
            return self.port.getInBusy()
        elif pinNumber == 12:
            return self.port.getInPaperOut()
        elif pinNumber == 13:
            return self.port.getInSelected()
        elif pinNumber == 15:
            return self.port.getInError()
        elif 2 <= pinNumber <= 9:
            return (self.port.PPRDATA() >> (pinNumber - 2)) & 1
        else:
            raise RuntimeError(
                f"Pin {pinNumber} cannot be read (by PParallelLinux.readPin())"
            )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..utils._checks import check_type
from ..utils._docs import copy_doc, fill_doc
from ..utils._imports import import_optional_dependency
from ._base import BaseTrigger

if TYPE_CHECKING:
    from mne_lsl.lsl import StreamInfo, StreamOutlet


@fill_doc
class LSLTrigger(BaseTrigger):
    """Trigger sending values on an LSL outlet.

    Make sure you are recording the stream created by the
    :class:`~stimuli.trigger.LSLTrigger` alongside your data. e.g. if you use
    LabRecorder, update the stream list after creating the
    :class:`~stimuli.trigger.LSLTrigger`.

    .. warning::

        Make sure to close the :class:`~mne_lsl.lsl.StreamOutlet` by calling the
        :meth:`~stimuli.trigger.LSLTrigger.close` method or by deleting the trigger
        after use.

    Parameters
    ----------
    name : str
        Name of the trigger displayed on the LSL network.

    Notes
    -----
    The :class:`~mne_lsl.lsl.StreamOutlet` created has the following properties:

    * Name: ``f"{name}"``
    * Type: ``"Markers"``
    * Number of channels: 1
    * Sampling rate: Irregular
    * Data type: ``np.int8``
    * Source ID: ``f"HNP-{name}"``

    The values sent must be in the range of strictly positive integers defined
    by ``np.int8``, 1 to 127 included.
    """

    def __init__(self, name: str) -> None:
        import_optional_dependency("mne_lsl")

        from mne_lsl.lsl import StreamInfo, StreamOutlet

        check_type(name, (str,), "name")
        self._name = name
        # create outlet
        self._sinfo = StreamInfo(
            name=name,
            stype="Markers",
            n_channels=1,
            sfreq=0.0,
            dtype="int8",
            source_id=f"HNP-{name}",
        )
        self._sinfo.set_channel_names(["STI"])
        self._sinfo.set_channel_types(["stim"])
        self._sinfo.set_channel_units(["none"])
        self._outlet = StreamOutlet(self._sinfo, max_buffered=1)

    @copy_doc(BaseTrigger.signal)
    def signal(self, value: int) -> None:
        value = super().signal(value)
        if not (1 <= value <= 127):
            raise ValueError(
                "The argument 'value' of an LSL trigger must be an integer "
                "between 1 and 127 included."
            )
        self._outlet.push_sample(np.array([value], dtype=np.int8))

    def close(self) -> None:
        """Close the LSL outlet."""
        if hasattr(self, "_outlet"):
            try:
                del self._outlet
            except Exception:  # pragma: no cover
                pass

    def __del__(self) -> None:  # noqa: D105
        self.close()

    # --------------------------------------------------------------------
    @property
    def name(self) -> str:
        """Name of the trigger displayed on the LSL network.

        :type: str
        """
        return self._name

    @property
    def sinfo(self) -> StreamInfo:
        """Description of the trigger outlet.

        :type: `~mne_lsl.lsl.StreamInfo`
        """
        return self._sinfo

    @property
    def outlet(self) -> StreamOutlet:
        """Trigger outlet.

        :type: `~mne_lsl.lsl.StreamOutlet`
        """
        return self._outlet



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.io import wavfile

from ..time import Clock
from ..utils._checks import ensure_path
from ..utils._docs import copy_doc, fill_doc
from ..utils.logs import warn
from ._base import BaseSound

if TYPE_CHECKING:
    from pathlib import Path

    import numpy as np
    from numpy.typing import NDArray

    from ..time import BaseClock

_SUPPORTED: tuple[str, ...] = (".wav",)


@fill_doc
class Sound(BaseSound):
    """Auditory stimulus loaded from a file.

    Parameters
    ----------
    fname : str | Path
        Path to the supported audio file to load.
    %(audio_device)s
    %(audio_backend)s
    %(audio_clock)s
    %(audio_kwargs)s
    """

    def __init__(
        self,
        fname: str | Path,
        device: int | None = None,
        *,
        backend: str = "sounddevice",
        clock: BaseClock = Clock,
        **kwargs,
    ) -> None:
        fname = ensure_path(fname, must_exist=True)
        if fname.suffix not in _SUPPORTED:
            raise ValueError(f"Unsupported file extension {fname.suffix}.")
        self._fname = fname
        sample_rate, original_signal = wavfile.read(self._fname)
        original_signal = original_signal.astype(np.float32)
        _check_signal(original_signal)
        volume = _extract_volume(original_signal)
        self._original_signal = _ensure_signal(original_signal)
        duration = self._original_signal.shape[0] / sample_rate
        super().__init__(
            volume,
            duration,
            sample_rate,
            device,
            self._original_signal.shape[1] if self._original_signal.ndim == 2 else 1,
            backend=backend,
            clock=clock,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Representation of the object."""
        return f"<Sound @ {self.fname}>"

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        super()._set_signal(self._original_signal)

    @BaseSound.duration.setter
    def duration(self, duration: float):  # noqa: D102
        warn(
            "The duration property of a loaded sound can not be changed. Skipping.",
        )

    @property
    def fname(self) -> Path:
        """The sound's original file name.

        :type: :class:`~pathlib.Path`
        """
        return self._fname


def _check_signal(signal: NDArray[np.float32]) -> None:
    """Check the loaded signal is valid."""
    if signal.ndim not in (1, 2):
        raise ValueError(
            "The signal must be a 1D array of shape (n_samples,) or a 2D array of "
            "shape (n_samples, n_channels)."
        )


def _extract_volume(signal: NDArray[np.float32]) -> NDArray[np.float32]:
    """Extract the volume from the signal."""
    if signal.ndim == 1:
        signal = signal[:, np.newaxis]  # add a channel dimension
    # normalize to retrieve the volume per channel
    max_ = np.max(np.abs(signal))
    if max_ == 0:
        raise RuntimeError("The loaded sound is silent.")
    signal /= max_
    volume = np.max(np.abs(signal, dtype=np.float32), axis=0) * 100
    assert any(elt == 100 for elt in volume)  # sanity-check
    return volume


def _ensure_signal(signal: NDArray[np.float32]) -> NDArray[np.float32]:
    """Ensure the signal is valid."""
    if signal.ndim == 2 and signal.shape[1] == 1:
        signal = signal.squeeze()
    signal /= np.max(np.abs(signal, dtype=np.float32), axis=0)
    np.nan_to_num(signal, copy=False, nan=0.0)  # sanity-check
    return signal



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from ..time import Clock
from ..utils._checks import check_type
from ..utils._docs import copy_doc, fill_doc
from ..utils.logs import logger
from ._base import BaseSound

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ..time import BaseClock


@fill_doc
class Tone(BaseSound):
    """Pure tone stimulus at a given frequency.

    Parameters
    ----------
    frequency : float
        Frequency of the tone in Hz.
    %(audio_volume)s
    %(audio_duration)s
    %(audio_sample_rate)s
    %(audio_device)s
    %(audio_n_channels)s
    %(audio_backend)s
    %(audio_clock)s
    %(audio_kwargs)s
    """

    def __init__(
        self,
        frequency: float,
        volume: float | Sequence[float],
        duration: float,
        sample_rate: int | None = None,
        device: int | None = None,
        n_channels: int = 1,
        *,
        backend: str = "sounddevice",
        clock: BaseClock = Clock,
        **kwargs,
    ) -> None:
        _check_frequency(frequency)
        self._frequency = frequency
        super().__init__(
            volume,
            duration,
            sample_rate,
            device,
            n_channels,
            backend=backend,
            clock=clock,
            **kwargs,
        )

    def __repr__(self) -> str:
        """Representation of the object."""
        return f"<Pure tone @ {self.frequency:.2f} Hz - {self.duration:.2f} s>"

    @copy_doc(BaseSound._set_signal)
    def _set_signal(self) -> None:
        signal = np.sin(2 * np.pi * self._frequency * self._times, dtype=np.float32)
        signal /= np.max(np.abs(signal))  # normalize
        signal = np.vstack([signal] * self._n_channels).T
        super()._set_signal(signal)

    @property
    def frequency(self) -> float:
        """The frequency of the tone in Hz.

        :type: :class:`float`
        """
        return self._frequency

    @frequency.setter
    def frequency(self, frequency: float) -> None:
        logger.debug("Setting 'frequency' to %.2f [Hz].", frequency)
        _check_frequency(frequency)
        self._frequency = frequency
        self._set_signal()


def _check_frequency(frequency: float) -> None:
    """Check that the frequency is valid."""
    check_type(frequency, ("numeric",), item_name="frequency")
    if frequency <= 0:
        raise ValueError(
            f"The frequency must be a strictly positive number. Provided '{frequency}' "
            "is invalid."
        )



"""Utility functions for checking types and values. Inspired from MNE."""

from __future__ import annotations

import logging
import operator
import os
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np

from ._docs import fill_doc

if TYPE_CHECKING:
    from typing import Any


def ensure_int(item: Any, item_name: str | None = None) -> int:
    """Ensure a variable is an integer.

    Parameters
    ----------
    item : Any
        Item to check.
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not int.
    """
    # This is preferred over numbers.Integral, see:
    # https://github.com/scipy/scipy/pull/7351#issuecomment-299713159
    try:
        # someone passing True/False is much more likely to be an error than
        # intentional usage
        if isinstance(item, bool):
            raise TypeError
        item = int(operator.index(item))
    except TypeError:
        item_name = "Item" if item_name is None else f"'{item_name}'"
        raise TypeError(f"{item_name} must be an integer, got {type(item)} instead.")

    return item


class _IntLike:
    @classmethod
    def __instancecheck__(cls, other: Any) -> bool:
        try:
            ensure_int(other)
        except TypeError:
            return False
        else:
            return True


class _Callable:
    @classmethod
    def __instancecheck__(cls, other: Any) -> bool:
        return callable(other)


_types = {
    "numeric": (np.floating, float, _IntLike()),
    "path-like": (str, Path, os.PathLike),
    "int-like": (_IntLike(),),
    "callable": (_Callable(),),
    "array-like": (list, tuple, set, np.ndarray),
}


def check_type(item: Any, types: tuple, item_name: str | None = None) -> None:
    """Check that item is an instance of types.

    Parameters
    ----------
    item : object
        Item to check.
    types : tuple of types | tuple of str
        Types to be checked against.
        If str, must be one of ('int-like', 'numeric', 'path-like', 'callable',
        'array-like').
    item_name : str | None
        Name of the item to show inside the error message.

    Raises
    ------
    TypeError
        When the type of the item is not one of the valid options.
    """
    check_types = sum(
        (
            (type(None),)
            if type_ is None
            else (type_,)
            if not isinstance(type_, str)
            else _types[type_]
            for type_ in types
        ),
        (),
    )

    if not isinstance(item, check_types):
        type_name = [
            "None"
            if cls_ is None
            else cls_.__name__
            if not isinstance(cls_, str)
            else cls_
            for cls_ in types
        ]
        if len(type_name) == 1:
            type_name = type_name[0]
        elif len(type_name) == 2:
            type_name = " or ".join(type_name)
        else:
            type_name[-1] = "or " + type_name[-1]
            type_name = ", ".join(type_name)
        item_name = "Item" if item_name is None else f"'{item_name}'"
        raise TypeError(
            f"{item_name} must be an instance of {type_name}, got {type(item)} instead."
        )


def check_value(
    item: Any,
    allowed_values: tuple | dict[Any, Any],
    item_name: str | None = None,
    extra: str | None = None,
) -> None:
    """Check the value of a parameter against a list of valid options.

    Parameters
    ----------
    item : object
        Item to check.
    allowed_values : tuple of objects | dict of objects
        Allowed values to be checked against. If a dictionary, checks against the keys.
    item_name : str | None
        Name of the item to show inside the error message.
    extra : str | None
        Extra string to append to the invalid value sentence, e.g. "when using DC mode".

    Raises
    ------
    ValueError
        When the value of the item is not one of the valid options.
    """
    if item not in allowed_values:
        item_name = "" if item_name is None else f" '{item_name}'"
        extra = "" if extra is None else " " + extra
        msg = (
            "Invalid value for the{item_name} parameter{extra}. "
            "{options}, but got {item!r} instead."
        )
        allowed_values = tuple(allowed_values)  # e.g., if a dict was given
        if len(allowed_values) == 1:
            options = f"The only allowed value is {repr(allowed_values[0])}"
        elif len(allowed_values) == 2:
            options = (
                f"Allowed values are {repr(allowed_values[0])} "
                f"and {repr(allowed_values[1])}"
            )
        else:
            options = "Allowed values are "
            options += ", ".join([f"{repr(v)}" for v in allowed_values[:-1]])
            options += f", and {repr(allowed_values[-1])}"
        raise ValueError(
            msg.format(item_name=item_name, extra=extra, options=options, item=item)
        )


@fill_doc
def check_verbose(verbose: Any) -> int:
    """Check that the value of verbose is valid.

    Parameters
    ----------
    %(verbose)s

    Returns
    -------
    verbose : int
        The verbosity level as an integer.
    """
    logging_types = dict(
        DEBUG=logging.DEBUG,
        INFO=logging.INFO,
        WARNING=logging.WARNING,
        ERROR=logging.ERROR,
        CRITICAL=logging.CRITICAL,
    )

    check_type(verbose, (bool, str, "int-like", None), item_name="verbose")

    if verbose is None:
        verbose = logging.WARNING
    elif isinstance(verbose, str):
        verbose = verbose.upper()
        check_value(verbose, logging_types, item_name="verbose")
        verbose = logging_types[verbose]
    elif isinstance(verbose, bool):
        if verbose:
            verbose = logging.INFO
        else:
            verbose = logging.WARNING
    elif isinstance(verbose, int):
        verbose = ensure_int(verbose)
        if verbose <= 0:
            raise ValueError(
                f"Argument 'verbose' can not be a negative integer, {verbose} is "
                "invalid."
            )

    return verbose


def ensure_path(item: Any, must_exist: bool) -> Path:
    """Ensure a variable is a Path.

    Parameters
    ----------
    item : Any
        Item to check.
    must_exist : bool
        If True, the path must resolve to an existing file or directory.

    Returns
    -------
    path : Path
        Path validated and converted to a pathlib.Path object.
    """
    try:
        item = Path(item)
    except TypeError:
        try:
            str_ = f"'{str(item)}' "
        except Exception:
            str_ = ""
        raise TypeError(
            f"The provided path {str_}is invalid and can not be converted. Please "
            f"provide a str, an os.PathLike or a pathlib.Path object, not {type(item)}."
        )
    if must_exist and not item.exists():
        raise FileNotFoundError(f"The provided path '{str(item)}' does not exist.")
    return item



"""Read / write access to the parallel port for Linux or Windows.

This code snippet is inspired from the parallel module of PsychoPy.
"""

import sys

# To make life easier, only try drivers which have a hope in heck of working. Because
# hasattr() in connection to windll ends up in an OSError trying to load 32bit drivers
# in a 64bit environment, different drivers defined in the dictionary 'drivers' are
# tested.

if sys.platform.startswith("linux"):
    from ._linux import PParallelLinux

    ParallelPort = PParallelLinux
elif sys.platform == "win32":
    drivers = dict(
        inpout32=("_inpout", "PParallelInpOut"),
        inpoutx64=("_inpout", "PParallelInpOut"),
        dlportio=("_dlportio", "PParallelDLPortIO"),
    )
    from ctypes import windll
    from importlib import import_module

    for key, val in drivers.items():
        driver_name, class_name = val
        try:
            hasattr(windll, key)
            ParallelPort = getattr(
                import_module("." + driver_name, __name__), class_name
            )
            break
        except (OSError, KeyError, NameError):
            ParallelPort = None
            continue
else:
    ParallelPort = None



"""Fill docstrings to avoid redundant docstrings in multiple files.

Inspired from mne: https://mne.tools/stable/index.html
Inspired from mne.utils.docs.py by Eric Larson <larson.eric.d@gmail.com>
"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any

# -- Documentation dictionary ----------------------------------------------------------
docdict: dict[str, str] = dict()

# -- A ---------------------------------------------------------------------------------
docdict["audio_backend"] = """
backend : ``"sounddevice"``
    The backend to use for sound playback."""

docdict["audio_clock"] = """
clock : BaseClock
    Clock object to use for timing measurements."""

docdict["audio_device"] = """
device : int | None
    Device index to use for sound playback. If None, the default device provided
    by the backend is used."""

docdict["audio_duration"] = """
duration : float
    Duration of the sound in seconds, used to generate the time array."""

docdict["audio_kwargs"] = """
**kwargs
    Additional keyword arguments passed to the backend initialization."""

docdict["audio_n_channels"] = """
n_channels : int
    Number of channels of the sound."""

docdict["audio_sample_rate"] = """
sample_rate : int | None
    Sample rate of the sound. If None, the default sample rate of the device
    provided by the backend is used."""

docdict["audio_volume"] = """
volume : float | list of float | tuple of float | array of float
    Volume of the sound as a percentage between 0 and 100. If a :class:`float` is
    provided, the same volume is used for all channels. If a sequence is provided,
    the length must match the number of channels."""

# -- B ---------------------------------------------------------------------------------
# -- C ---------------------------------------------------------------------------------
# -- D ---------------------------------------------------------------------------------
# -- E ---------------------------------------------------------------------------------
# -- F ---------------------------------------------------------------------------------
# -- G ---------------------------------------------------------------------------------
# -- H ---------------------------------------------------------------------------------
# -- I ---------------------------------------------------------------------------------
# -- J ---------------------------------------------------------------------------------
# -- K ---------------------------------------------------------------------------------
# -- L ---------------------------------------------------------------------------------
# -- M ---------------------------------------------------------------------------------
# -- N ---------------------------------------------------------------------------------
# -- O ---------------------------------------------------------------------------------
# -- P ---------------------------------------------------------------------------------
# -- Q ---------------------------------------------------------------------------------
# -- R ---------------------------------------------------------------------------------
# -- S ---------------------------------------------------------------------------------
# -- T ---------------------------------------------------------------------------------
# -- U ---------------------------------------------------------------------------------
# -- V ---------------------------------------------------------------------------------
docdict["verbose"] = """
verbose : int | str | bool | None
    Sets the verbosity level. The verbosity increases gradually between ``"CRITICAL"``,
    ``"ERROR"``, ``"WARNING"``, ``"INFO"`` and ``"DEBUG"``. If None is provided, the
    verbosity is set to ``"WARNING"``. If a bool is provided, the verbosity is set to
    ``"WARNING"`` for False and to ``"INFO"`` for True."""

# -- W ---------------------------------------------------------------------------------
# -- X ---------------------------------------------------------------------------------
# -- Y ---------------------------------------------------------------------------------
# -- Z ---------------------------------------------------------------------------------

# -- Documentation functions -----------------------------------------------------------
docdict_indented: dict[int, dict[str, str]] = dict()


def fill_doc(f: Callable[..., Any]) -> Callable[..., Any]:
    """Fill a docstring with docdict entries.

    Parameters
    ----------
    f : callable
        The function to fill the docstring of (modified in place).

    Returns
    -------
    f : callable
        The function, potentially with an updated __doc__.
    """
    docstring = f.__doc__
    if not docstring:
        return f

    lines = docstring.splitlines()
    indent_count = _indentcount_lines(lines)

    try:
        indented = docdict_indented[indent_count]
    except KeyError:
        indent = " " * indent_count
        docdict_indented[indent_count] = indented = dict()

        for name, docstr in docdict.items():
            lines = [
                indent + line if k != 0 else line
                for k, line in enumerate(docstr.strip().splitlines())
            ]
            indented[name] = "\n".join(lines)

    try:
        f.__doc__ = docstring % indented
    except (TypeError, ValueError, KeyError) as exp:
        funcname = f.__name__
        funcname = docstring.split("\n")[0] if funcname is None else funcname
        raise RuntimeError(f"Error documenting {funcname}:\n{str(exp)}")

    return f


def _indentcount_lines(lines: list[str]) -> int:
    """Minimum indent for all lines in line list.

    >>> lines = [" one", "  two", "   three"]
    >>> indentcount_lines(lines)
    1
    >>> lines = []
    >>> indentcount_lines(lines)
    0
    >>> lines = [" one"]
    >>> indentcount_lines(lines)
    1
    >>> indentcount_lines(["    "])
    0
    """
    indent = sys.maxsize
    for k, line in enumerate(lines):
        if k == 0:
            continue
        line_stripped = line.lstrip()
        if line_stripped:
            indent = min(indent, len(line) - len(line_stripped))
    return indent


def copy_doc(source: Callable[..., Any]) -> Callable[..., Any]:
    """Copy the docstring from another function (decorator).

    The docstring of the source function is prepepended to the docstring of the function
    wrapped by this decorator.

    This is useful when inheriting from a class and overloading a method. This decorator
    can be used to copy the docstring of the original method.

    Parameters
    ----------
    source : callable
        The function to copy the docstring from.

    Returns
    -------
    wrapper : callable
        The decorated function.

    Examples
    --------
    >>> class A:
    ...     def m1():
    ...         '''Docstring for m1'''
    ...         pass
    >>> class B(A):
    ...     @copy_doc(A.m1)
    ...     def m1():
    ...         '''this gets appended'''
    ...         pass
    >>> print(B.m1.__doc__)
    Docstring for m1 this gets appended
    """

    def wrapper(func: Callable[..., Any]) -> Callable[..., Any]:
        if source.__doc__ is None or len(source.__doc__) == 0:
            raise RuntimeError(
                f"The docstring from {source.__name__} could not be copied because it "
                "was empty."
            )
        doc = source.__doc__
        if func.__doc__ is not None:
            doc += func.__doc__
        func.__doc__ = doc
        return func

    return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING

from . import sounddevice
from .sounddevice import SoundSD

if TYPE_CHECKING:
    from ._base import BaseBackend

BACKENDS: dict[str, BaseBackend] = {"sounddevice": SoundSD}



from __future__ import annotations

from typing import TYPE_CHECKING

import sounddevice as sd

from ...time import Clock, sleep
from ...utils._checks import check_value, ensure_int
from ...utils._docs import copy_doc
from ...utils.logs import warn
from ._base import BaseBackend

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from ...time import BaseClock


class SoundSD(BaseBackend):
    """Sounddevice backend for audio playback.

    Parameters
    ----------
    device : int
        Device index of the output device as provided by
        :func:`sounddevice.query_devices()`. If None, the default output device is used.
    sample_rate : int
        The sample rate of the audio data, which should match the sample rate of the
        output device. If None, the default sample rate of the device is used.
    clock : BaseClock
        Clock object to use for time measurement. By default, the
        :class:`stimuli.time.Clock` class is used.
    """

    def __init__(
        self,
        device: int | None,
        sample_rate: int | None,
        *,
        clock: BaseClock = Clock,
    ) -> None:
        super().__init__(sample_rate, device, clock)
        self._device = _ensure_device(device)
        self._sample_rate = _ensure_sample_rate(sample_rate, self._device)
        if self._sample_rate != self._device["default_samplerate"]:
            warn(
                "The sample rate provided to the 'SoundSD' backend "
                f"({self._sample_rate}) differs from the default sample rate of the "
                f"device ({self._device['default_samplerate']})."
            )

    @copy_doc(BaseBackend.close)
    def close(self) -> None:
        if hasattr(self, "_target_time") and self._target_time is not None:
            warn("The audio playback was on-going.")
        if hasattr(self, "_stream"):
            self._stream.stop()
            self._stream.close()

    @copy_doc(BaseBackend.initialize)
    def initialize(self, data: NDArray, block_size: int = 32) -> None:
        # fmt: off
        """block_size : int
            The number of frames passed to the stream callback function, or the
            preferred block granularity for a blocking read/write stream. The special
            value ``blocksize=0`` may be used to request that the stream callback will
            receive an optimal (and possibly varying) number of frames based on host
            requirements and the requested latency settings.
        """
        # fmt: on
        super().initialize(data)
        if (
            self._data.ndim == 2
            and self._device["max_output_channels"] < self._data.shape[1]
        ):
            raise ValueError(
                f"Device '{self._device['index']}: {self._device['name']}' does not "
                f"support the number of output channels ({self._data.shape[1]})."
            )
        block_size = _ensure_block_size(block_size)
        # convert the data array to a supported byte representation
        check_value(self._data.dtype.name, sd._sampleformats, "dtype")
        self._bytes_per_frame = self._data.shape[1] * self._data.itemsize
        n_channels = self._data.shape[1]
        dtype = self._data.dtype
        self._data = self._data.tobytes(order="C")
        # store callback variables
        self._current_frame = 0
        self._target_time = None
        # create and open the output stream
        self._stream = sd.RawOutputStream(
            blocksize=block_size,
            callback=self._callback,
            channels=n_channels,
            device=self._device["index"],
            dtype=dtype,
            latency="low",
            samplerate=self._sample_rate,
            prime_output_buffers_using_stream_callback=True,
        )
        self._stream.start()

    def _callback(self, outdata, frames, time_info, status):
        """Callback audio function."""  # noqa: D401
        if self._target_time is None:
            outdata[:] = b"\x00" * frames * self._bytes_per_frame
            return
        delta_ns = int((time_info.outputBufferDacTime - time_info.currentTime) * 1e9)
        if self._clock.get_time_ns() + delta_ns < self._target_time:
            outdata[:] = b"\x00" * frames * self._bytes_per_frame
            return
        start = self._current_frame * self._bytes_per_frame
        end = start + frames * self._bytes_per_frame
        if end <= len(self._data):
            outdata[:] = self._data[start:end]
            self._current_frame += frames
        else:
            remaining_bytes = len(self._data) - start
            outdata[:remaining_bytes] = self._data[start:]
            outdata[remaining_bytes:] = b"\x00" * (end - len(self._data))
            # reset for the next playback
            self._current_frame = 0
            self._target_time = None

    @copy_doc(BaseBackend.play)
    def play(self, when: float | None = None, *, blocking: bool = False) -> None:
        super().play(when, blocking=blocking)
        if self._target_time is not None:
            raise RuntimeError("The audio playback is already on-going.")
        self._target_time = (
            self._clock.get_time_ns()
            if when is None
            else self._clock.get_time_ns() + int(when * 1e9)
        )
        if blocking:
            wait = self._duration
            if when is not None:
                wait += when
            sleep(0.95 * wait)  # 5% margin
            while self._target_time is not None:
                # hog CPU for the last blocking part, based on _target_time and not on
                # an estimated duration as sleep(...) does.
                pass

    @copy_doc(BaseBackend.stop)
    def stop(self) -> None:
        super().stop()
        if self._target_time is None:
            warn("The audio playback was not on-going.")
        self._target_time = None


def _ensure_block_size(block_size: int) -> int:
    """Ensure the block_size argument is valid."""
    block_size = ensure_int(block_size, "block_size")
    if block_size < 0:
        raise ValueError(
            f"Argument 'block_size' must be greater or equal than 0. '{block_size}' is "
            "invalid."
        )
    return block_size


def _ensure_device(device: int | None) -> dict[str, str | int | float]:
    """Ensure the device argument is valid."""
    if device is None:
        idx = sd.default.device["output"]
        return sd.query_devices()[idx]
    device_idx = ensure_int(device, "device")
    devices = sd.query_devices()
    if len(devices) <= device_idx:
        raise ValueError(
            f"Invalid device index. There are only {len(devices)} devices."
        )
    device = devices[device_idx]
    if device["max_output_channels"] <= 0:
        raise ValueError(
            f"Device '{device_idx}: {device['name']}' does not support output "
            "channels. Please select a different device."
        )
    assert device_idx == device["index"]  # sanity-check
    return device


def _ensure_sample_rate(
    sample_rate: int | None, device: dict[str, str | int | float]
) -> int:
    """Ensure the sample_rate argument is valid."""
    if sample_rate is None:
        return int(device["default_samplerate"])
    sample_rate = ensure_int(sample_rate, "sample_rate")
    if sample_rate <= 0:
        raise ValueError(
            f"Argument 'sample_rate' must be greater than 0. '{sample_rate}' "
            "is invalid."
        )
    return sample_rate
